@@ -33,8 +33,7 @@
 #include "absl/strings/strip.h"
 #include "absl/types/optional.h"
 
-#include <grpc/event_engine/event_engine.h>
-#include <grpc/grpc.h>
+#include <grpc/impl/channel_arg_names.h>
 #include <grpc/support/alloc.h>
 #include <grpc/support/log.h>
 
@@ -61,7 +60,6 @@
 
 #include <address_sorting/address_sorting.h>
 
-#include "absl/container/flat_hash_set.h"
 #include "absl/strings/str_cat.h"
 
 #include "src/core/ext/filters/client_channel/lb_policy/grpclb/grpclb_balancer_addresses.h"
@@ -70,9 +68,8 @@
 #include "src/core/lib/backoff/backoff.h"
 #include "src/core/lib/channel/channel_args.h"
 #include "src/core/lib/config/config_vars.h"
-#include "src/core/lib/event_engine/handle_containers.h"
 #include "src/core/lib/iomgr/resolve_address.h"
-#include "src/core/lib/resolver/server_address.h"
+#include "src/core/lib/resolver/endpoint_addresses.h"
 #include "src/core/lib/service_config/service_config_impl.h"
 #include "src/core/lib/transport/error_utils.h"
 
@@ -180,9 +177,9 @@ class AresClientChannelDNSResolver : public PollingResolver {
     std::unique_ptr<grpc_ares_request> txt_request_
         ABSL_GUARDED_BY(on_resolved_mu_);
     // Output fields from ares request.
-    std::unique_ptr<ServerAddressList> addresses_
+    std::unique_ptr<EndpointAddressesList> addresses_
         ABSL_GUARDED_BY(on_resolved_mu_);
-    std::unique_ptr<ServerAddressList> balancer_addresses_
+    std::unique_ptr<EndpointAddressesList> balancer_addresses_
         ABSL_GUARDED_BY(on_resolved_mu_);
     char* service_config_json_ ABSL_GUARDED_BY(on_resolved_mu_) = nullptr;
   };
@@ -227,7 +224,8 @@ AresClientChannelDNSResolver::~AresClientChannelDNSResolver() {
 
 OrphanablePtr<Orphanable> AresClientChannelDNSResolver::StartRequest() {
   return MakeOrphanable<AresRequestWrapper>(
-      Ref(DEBUG_LOCATION, "dns-resolving"));
+      RefAsSubclass<AresClientChannelDNSResolver>(DEBUG_LOCATION,
+                                                  "dns-resolving"));
 }
 
 void AresClientChannelDNSResolver::AresRequestWrapper::OnHostnameResolved(
@@ -301,7 +299,7 @@ AresClientChannelDNSResolver::AresRequestWrapper::OnResolvedLocked(
     if (addresses_ != nullptr) {
       result.addresses = std::move(*addresses_);
     } else {
-      result.addresses = ServerAddressList();
+      result.addresses.emplace();
     }
     if (service_config_json_ != nullptr) {
       auto service_config_string = ChooseServiceConfig(service_config_json_);
@@ -322,8 +320,8 @@ AresClientChannelDNSResolver::AresRequestWrapper::OnResolvedLocked(
       }
     }
     if (balancer_addresses_ != nullptr) {
-      result.args = SetGrpcLbBalancerAddresses(
-          result.args, ServerAddressList(*balancer_addresses_));
+      result.args =
+          SetGrpcLbBalancerAddresses(result.args, *balancer_addresses_);
     }
   } else {
     GRPC_CARES_TRACE_LOG("resolver:%p dns resolution failed: %s", this,
@@ -537,7 +535,7 @@ class AresDNSResolver : public DNSResolver {
         absl::StatusOr<std::vector<grpc_resolved_address>>)>
         on_resolve_address_done_;
     // currently resolving addresses
-    std::unique_ptr<ServerAddressList> addresses_;
+    std::unique_ptr<EndpointAddressesList> addresses_;
   };
 
   class AresSRVRequest : public AresRequest {
@@ -585,7 +583,7 @@ class AresDNSResolver : public DNSResolver {
         absl::StatusOr<std::vector<grpc_resolved_address>>)>
         on_resolve_address_done_;
     // currently resolving addresses
-    std::unique_ptr<ServerAddressList> balancer_addresses_;
+    std::unique_ptr<EndpointAddressesList> balancer_addresses_;
   };
 
   class AresTXTRequest : public AresRequest {
@@ -706,8 +704,7 @@ class AresDNSResolver : public DNSResolver {
   // the previous default DNS resolver, used to delegate blocking DNS calls to
   std::shared_ptr<DNSResolver> default_resolver_ = GetDNSResolver();
   Mutex mu_;
-  grpc_event_engine::experimental::LookupTaskHandleSet open_requests_
-      ABSL_GUARDED_BY(mu_);
+  TaskHandleSet open_requests_ ABSL_GUARDED_BY(mu_);
   intptr_t aba_token_ ABSL_GUARDED_BY(mu_) = 0;
 };
 
