@@ -14,8 +14,8 @@
 // limitations under the License.
 //
 
-#ifndef GRPC_CORE_EXT_XDS_XDS_CLIENT_H
-#define GRPC_CORE_EXT_XDS_XDS_CLIENT_H
+#ifndef GRPC_SRC_CORE_EXT_XDS_XDS_CLIENT_H
+#define GRPC_SRC_CORE_EXT_XDS_XDS_CLIENT_H
 
 #include <grpc/support/port_platform.h>
 
@@ -30,7 +30,7 @@
 #include "absl/status/status.h"
 #include "absl/status/statusor.h"
 #include "absl/strings/string_view.h"
-#include "upb/def.hpp"
+#include "upb/reflection/def.hpp"
 
 #include <grpc/event_engine/event_engine.h>
 
@@ -56,6 +56,11 @@ extern TraceFlag grpc_xds_client_refcount_trace;
 
 class XdsClient : public DualRefCounted<XdsClient> {
  public:
+  class ReadDelayHandle : public RefCounted<ReadDelayHandle> {
+   public:
+    static RefCountedPtr<ReadDelayHandle> NoWait() { return nullptr; }
+  };
+
   // Resource watcher interface.  Implemented by callers.
   // Note: Most callers will not use this API directly but rather via a
   // resource-type-specific wrapper API provided by the relevant
@@ -63,11 +68,14 @@ class XdsClient : public DualRefCounted<XdsClient> {
   class ResourceWatcherInterface : public RefCounted<ResourceWatcherInterface> {
    public:
     virtual void OnGenericResourceChanged(
-        const XdsResourceType::ResourceData* resource)
+        std::shared_ptr<const XdsResourceType::ResourceData> resource,
+        RefCountedPtr<ReadDelayHandle> read_delay_handle)
         ABSL_EXCLUSIVE_LOCKS_REQUIRED(&work_serializer_) = 0;
-    virtual void OnError(absl::Status status)
+    virtual void OnError(absl::Status status,
+                         RefCountedPtr<ReadDelayHandle> read_delay_handle)
         ABSL_EXCLUSIVE_LOCKS_REQUIRED(&work_serializer_) = 0;
-    virtual void OnResourceDoesNotExist()
+    virtual void OnResourceDoesNotExist(
+        RefCountedPtr<ReadDelayHandle> read_delay_handle)
         ABSL_EXCLUSIVE_LOCKS_REQUIRED(&work_serializer_) = 0;
   };
 
@@ -171,23 +179,23 @@ class XdsClient : public DualRefCounted<XdsClient> {
 
   // Contains a channel to the xds server and all the data related to the
   // channel.  Holds a ref to the xds client object.
-  class ChannelState : public DualRefCounted<ChannelState> {
+  class XdsChannel : public DualRefCounted<XdsChannel> {
    public:
     template <typename T>
     class RetryableCall;
 
-    class AdsCallState;
-    class LrsCallState;
+    class AdsCall;
+    class LrsCall;
 
-    ChannelState(WeakRefCountedPtr<XdsClient> xds_client,
-                 const XdsBootstrap::XdsServer& server);
-    ~ChannelState() override;
+    XdsChannel(WeakRefCountedPtr<XdsClient> xds_client,
+               const XdsBootstrap::XdsServer& server);
+    ~XdsChannel() override;
 
     void Orphan() override;
 
     XdsClient* xds_client() const { return xds_client_.get(); }
-    AdsCallState* ads_calld() const;
-    LrsCallState* lrs_calld() const;
+    AdsCall* ads_call() const;
+    LrsCall* lrs_call() const;
 
     void ResetBackoff();
 
@@ -223,9 +231,9 @@ class XdsClient : public DualRefCounted<XdsClient> {
 
     bool shutting_down_ = false;
 
-    // The retryable XDS calls.
-    OrphanablePtr<RetryableCall<AdsCallState>> ads_calld_;
-    OrphanablePtr<RetryableCall<LrsCallState>> lrs_calld_;
+    // The retryable ADS and LRS calls.
+    OrphanablePtr<RetryableCall<AdsCall>> ads_call_;
+    OrphanablePtr<RetryableCall<LrsCall>> lrs_call_;
 
     // Stores the most recent accepted resource version for each resource type.
     std::map<const XdsResourceType*, std::string /*version*/>
@@ -238,13 +246,13 @@ class XdsClient : public DualRefCounted<XdsClient> {
     std::map<ResourceWatcherInterface*, RefCountedPtr<ResourceWatcherInterface>>
         watchers;
     // The latest data seen for the resource.
-    std::unique_ptr<XdsResourceType::ResourceData> resource;
+    std::shared_ptr<const XdsResourceType::ResourceData> resource;
     XdsApi::ResourceMetadata meta;
     bool ignored_deletion = false;
   };
 
   struct AuthorityState {
-    RefCountedPtr<ChannelState> channel_state;
+    RefCountedPtr<XdsChannel> xds_channel;
     std::map<const XdsResourceType*, std::map<XdsResourceKey, ResourceState>>
         resource_map;
   };
@@ -269,7 +277,7 @@ class XdsClient : public DualRefCounted<XdsClient> {
       LoadReportState>;
 
   struct LoadReportServer {
-    RefCountedPtr<ChannelState> channel_state;
+    RefCountedPtr<XdsChannel> xds_channel;
     LoadReportMap load_report_map;
   };
 
@@ -277,11 +285,12 @@ class XdsClient : public DualRefCounted<XdsClient> {
   void NotifyWatchersOnErrorLocked(
       const std::map<ResourceWatcherInterface*,
                      RefCountedPtr<ResourceWatcherInterface>>& watchers,
-      absl::Status status);
+      absl::Status status, RefCountedPtr<ReadDelayHandle> read_delay_handle);
   // Sends a resource-does-not-exist notification to a specific set of watchers.
   void NotifyWatchersOnResourceDoesNotExist(
       const std::map<ResourceWatcherInterface*,
-                     RefCountedPtr<ResourceWatcherInterface>>& watchers);
+                     RefCountedPtr<ResourceWatcherInterface>>& watchers,
+      RefCountedPtr<ReadDelayHandle> read_delay_handle);
 
   void MaybeRegisterResourceTypeLocked(const XdsResourceType* resource_type)
       ABSL_EXCLUSIVE_LOCKS_REQUIRED(mu_);
@@ -300,7 +309,7 @@ class XdsClient : public DualRefCounted<XdsClient> {
       const XdsBootstrap::XdsServer& xds_server, bool send_all_clusters,
       const std::set<std::string>& clusters) ABSL_EXCLUSIVE_LOCKS_REQUIRED(mu_);
 
-  RefCountedPtr<ChannelState> GetOrCreateChannelStateLocked(
+  RefCountedPtr<XdsChannel> GetOrCreateXdsChannelLocked(
       const XdsBootstrap::XdsServer& server, const char* reason)
       ABSL_EXCLUSIVE_LOCKS_REQUIRED(mu_);
 
@@ -317,12 +326,12 @@ class XdsClient : public DualRefCounted<XdsClient> {
   // Stores resource type objects seen by type URL.
   std::map<absl::string_view /*resource_type*/, const XdsResourceType*>
       resource_types_ ABSL_GUARDED_BY(mu_);
-  upb::SymbolTable symtab_ ABSL_GUARDED_BY(mu_);
+  upb::DefPool def_pool_ ABSL_GUARDED_BY(mu_);
 
   // Map of existing xDS server channels.
   // Key is owned by the bootstrap config.
-  std::map<const XdsBootstrap::XdsServer*, ChannelState*>
-      xds_server_channel_map_ ABSL_GUARDED_BY(mu_);
+  std::map<const XdsBootstrap::XdsServer*, XdsChannel*> xds_server_channel_map_
+      ABSL_GUARDED_BY(mu_);
 
   std::map<std::string /*authority*/, AuthorityState> authority_state_map_
       ABSL_GUARDED_BY(mu_);
@@ -341,4 +350,4 @@ class XdsClient : public DualRefCounted<XdsClient> {
 
 }  // namespace grpc_core
 
-#endif  // GRPC_CORE_EXT_XDS_XDS_CLIENT_H
+#endif  // GRPC_SRC_CORE_EXT_XDS_XDS_CLIENT_H
